# The function that validates a MassBank record
# 
#' MassBank record creation workflow
#' 
#' Uses data generated by \code{\link{msmsWorkflow}} to create MassBank records.
#' 
#' See the vignette \code{vignette("RMassBank")} for detailed informations about the usage.
#' 
#' Steps:
#' 
#' Step 1: Find which compounds don't have annotation information yet. For these
#' 		 compounds, pull information from CTS (using gatherData).
#' 
#' Step 2: If new compounds were found, then export the infolist.csv and stop the workflow.
#' 		Otherwise, continue.
#' 
#' Step 3: Take the archive data (in table format) and reformat it to MassBank tree format.
#' 
#' Step 4: Compile the spectra. Using the skeletons from the archive data, create
#'   MassBank records per compound and fill them with peak data for each spectrum.
#'   Also, assign accession numbers based on scan mode and relative scan no.
#' 
#' Step 5: Convert the internal tree-like representation of the MassBank data into
#'  flat-text string arrays (basically, into text-file style, but still in memory)
#' 
#' Step 6: For all OK records, generate a corresponding molfile with the structure
#'   of the compound, based on the SMILES entry from the MassBank record. (This molfile
#'   is still in memory only, not yet a physical file)
#' 
#' Step 7: If necessary, generate the appropriate subdirectories, and actually write
#'   the files to disk.
#' 
#' Step 8: Create the list.tsv in the molfiles folder, which is required by MassBank
#'   to attribute substances to their corresponding structure molfiles. 
#' 
#' @param Files A path to the records that should be validated. Can be a directory with multiple files.
#' @return Nothing
#' @seealso \code{\link{mbWorkspace-class}}
#' @author Somebody
#' @examples \dontrun{
#' 		mb <- newMbWorkspace(w) # w being a msmsWorkspace
#' 		mb <- loadInfolists(mb, "D:/myInfolistPath")
#' 		mb <- mbWorkflow(mb, steps=c(1:3), "newinfos.csv")
#' 		
#' }
#' @export
validate <- function(File){
	##This Validator uses the syntax of RMassBank
	Instrument_List <<- .getInstruments()
	mb <<- new("mbWorkspace")
	testNumber <<- new("numeric")
	ReadAnnotations <<- FALSE
	##Is the argument a directory?
	##If yes, 
	if(file.info(File[1])$isdir){
	    Files <- list.files(path = File, full.names = TRUE)
	}
	
	##I need to do this globally, because RUnit-Tests don't accept environments.
	testNumber <<- 1
	
	##Parsing with the help of an R function
	mb <<- parseMassBank(Files)

	## Test RMassBank Objects
	tests <- list()

	for(i in 1:length(Files)){
		if(mb@compiled_ok[[i]][['AC$MASS_SPECTROMETRY']][['MS_TYPE']] == "MS2" || mb@compiled_ok[[i]][['AC$MASS_SPECTROMETRY']][['MS_TYPE']] == "MS"){
		tests[[i]] <- defineTestSuite(Files[i], dirs = system.file(package="RMassBank", "unitTests"), testFileRegexp = "^runit.MS2.test.[rR]$",
                testFuncRegexp = "^test.+",
                rngKind = "Marsaglia-Multicarry",
                rngNormalKind = "Kinderman-Ramage")
		} else{
			tests[[i]] <- defineTestSuite(Files[i], dirs = system.file(package="RMassBank", "unitTests"), testFileRegexp = "^runit.MSn.test.[rR]$",
                testFuncRegexp = "^test.+",
                rngKind = "Marsaglia-Multicarry",
                rngNormalKind = "Kinderman-Ramage")
		}
	}
	print("Starting Tests")
	flush.console()
	testData <- runTestSuite(tests)
	printHTMLProtocol(testData, fileName = paste(system.file(package = "RMassBank"),"report.html", sep = ""))
	print(paste("Report for the file(s) finished"))
}

##Most of these functions will be shorter(and there'll probably be one less) when I found the solution for a bug

##This function checks if an .obo-file is readable for ontoCAT
.isOboReadable <- function(filename){

	##getOntology() has a problem with reading relative Windows paths(it wants an URI),
	##so the path has to be made absolute
	##I reckon this should work under Linux without doing that
	ont <- getOntology(normalizePath(filename))
	if(is.null(getOntologyAccession(ont))){
		return(FALSE)
	}
	return(TRUE)
}

##This function downloads the psi-ms.obo-ontology so we can get the allowed instrument-names
##This is a _temporary_ fix until I find out why getOntology() doesn't work when there are "import:"-lines in the .obo-file
##Until then I will simply remove them, because we don't need the imported ontologies
.downloadPsiObo <- function(){
		connPsiObo <- url("http://psidev.cvs.sourceforge.net/viewvc/psidev/psi/psi-ms/mzML/controlledVocabulary/psi-ms.obo")
		oboFile <- readLines(connPsiObo)
		close(connPsiObo)
		oboFile <- oboFile[-grep("import:",oboFile)] 
		connLocal <- file("psi-ms.obo")
		writeLines(oboFile,connLocal)
		close(connLocal)
}

##Checks if the psi-ms.obo is there
##Will be converted to "checkforinstruments" as soon as I can find the problem
##with getOntology()
.checkForPsiMs <- function(){
	
	if(file.exists("psi-ms.obo")){
		if(.isOboReadable("psi-ms.obo")){
			print("It seems that you have a working psi-ms.obo, do you want to update it? [y/n]")
			while(TRUE){
				answer <- readLines(stdin(), n=1, warn=FALSE)
				if(answer == "y"){
					.downloadPsiObo()
					return(TRUE)
				}
				if(answer == "n"){
					return(TRUE)
				}
				print("Please type exactly y or n")
			}
		}
	}
	.downloadPsiObo()
	return(TRUE)
}

##This is a list of the possible instrument names 
.getInstruments <- function(){
	Onto <- getOntology(system.file(package = "RMassBank", "psi-ms.obo"))
	instrumentTerms <- getAllTermChildrenById(Onto,"MS_1000031")
	instruments <- vector()	
	for(i in 1:length(instrumentTerms)){
		instruments[i] <- getLabel(instrumentTerms[[i]])
	}
	return(instruments)
}

.smiles2mass <- function(SMILES){
	massfromformula <- parse.smiles(SMILES)[[1]]
	do.typing(massfromformula)
	do.aromaticity(massfromformula)
	convert.implicit.to.explicit(massfromformula)
	do.isotopes(massfromformula)
	mass <- get.exact.mass(massfromformula)
	return(mass)
}